import base64
import gc
import io
import json
import os
from typing import List
from urllib.error import HTTPError
from urllib.parse import quote
from urllib.request import urlopen

import extra_streamlit_components as stx
import requests
import streamlit as st
import streamlit.components.v1 as components
from better_profanity import profanity
from docarray import Document, DocumentArray
from docarray.typing import Image, Text
from streamlit.runtime import Runtime
from streamlit.runtime.scriptrunner import add_script_run_ctx
from streamlit.web.server.server import Server
from tornado.httputil import parse_cookie

from now.app.base.create_jcloud_name import create_jcloud_name
from now.constants import MODALITY_TO_MODELS, NOWGATEWAY_BFF_PORT
from now.executor.gateway.playground.src.constants import (
    BUTTONS,
    S3_DEMO_PATH,
    SSO_COOKIE,
    ds_set,
)
from now.executor.gateway.playground.src.search import (
    call_flow,
    get_query_params,
    multimodal_search,
)
from now.now_dataclasses import UserInput

dir_path = os.path.dirname(os.path.realpath(__file__))

# HEADER
st.set_page_config(page_title='NOW', page_icon='https://jina.ai/favicon.ico')
profanity.load_censor_words()


def convert_file_to_document(query):
    data = query.read()
    doc = Document(blob=data)
    return doc


@st.cache(allow_output_mutation=True, suppress_st_warning=True)
def get_cookie_manager():
    return stx.CookieManager()


cookie_manager = get_cookie_manager()


def _get_all_cookies() -> dict:
    session_id = add_script_run_ctx().streamlit_script_run_ctx.session_id
    session_info = Server.get_current()._get_session_info(session_id)
    header = session_info.ws.request.headers
    cookie_strings = [header_str for k, header_str in header.get_all() if k == 'Cookie']
    parsed_cookies = {k: v for c in cookie_strings for k, v in parse_cookie(c).items()}

    return parsed_cookies


def get_cookie_value(cookie_name):
    all_cookies = _get_all_cookies()
    for k, v in all_cookies.items():
        if k == cookie_name:
            return v


def nav_to(url):
    nav_script = """
        <meta http-equiv="refresh" content="0; url='%s'">
    """ % (
        url
    )
    st.write(nav_script, unsafe_allow_html=True)


def toggle_score_breakdown():
    if st.checkbox('Show score breakdown', key='scores'):
        st.session_state.show_score_breakdown = True
    else:
        st.session_state.show_score_breakdown = False


def deploy_streamlit(user_input: UserInput):
    """
    We want to provide the end-to-end experience to the user.
    Please deploy a streamlit playground on k8s/local to access the api.
    You can get the starting point for the streamlit application from alex.
    """
    # Start with setting up the vars default values then proceed to placing UI components
    # Set up session state vars if not already set
    setup_session_state()

    # Retrieve query params
    params = get_query_params()
    # update query params with user input
    for key, val in user_input.dict().items():
        setattr(params, key, val)
    redirect_to = render_auth_components(params)

    _, mid, _ = st.columns([0.8, 1, 1])
    with open(os.path.join(dir_path, 'logo.svg'), 'r') as f:
        svg = f.read()
    with mid:
        b64 = base64.b64encode(svg.encode('utf-8')).decode('utf-8')
        html = r'<img width="250" src="data:image/svg+xml;base64,%s"/>' % b64
        st.write(html, unsafe_allow_html=True)

    if redirect_to and st.session_state.login:
        nav_to(redirect_to)
    else:
        da_img, da_txt = load_example_queries(params.data)

        setup_design()

        if st.session_state.filters == 'notags':
            try:
                tags = get_info_from_endpoint(params, endpoint='tags')
                st.session_state.filters = tags
            except Exception as e:
                print("Filters couldn't be loaded from the endpoint properly.", e)
                st.session_state.filters = 'notags'
        if not st.session_state.index_fields_dict:
            st.session_state.index_fields_dict = get_info_from_endpoint(
                params, 'encoder_to_dataclass_fields_mods'
            )
            st.session_state.field_names_to_dataclass_fields = (
                user_input.field_names_to_dataclass_fields
            )

        filter_selection = {}
        if st.session_state.filters != 'notags':
            st.sidebar.title('Filters')
            if not st.session_state.filters_set:
                for tag, values in st.session_state.filters.items():
                    values.insert(0, 'All')
                    filter_selection[tag] = st.sidebar.selectbox(tag, values)
                st.session_state.filters_set = True
            else:
                for tag, values in st.session_state.filters.items():
                    filter_selection[tag] = st.sidebar.selectbox(tag, values)

        if st.session_state.filters != 'notags' and not st.session_state.filters_set:
            st.sidebar.title('Filters')
            for tag, values in st.session_state.filters.items():
                values.insert(0, 'All')
                filter_selection[tag] = st.sidebar.selectbox(tag, values)
        l, r = st.columns([5, 5])
        with l:
            st.header('Text')
            render_mm_query(st.session_state['query'], 'text')
        with r:
            st.header('Image')
            render_mm_query(st.session_state['query'], 'image')

        customize_semantic_scores()
        toggle_score_breakdown()
        search_mapping_list = list(st.session_state['query'].values())
        if any([d['value'] for d in search_mapping_list]):
            st.session_state.matches = multimodal_search(
                query_field_values_modalities=list(
                    filter(lambda x: x['value'], search_mapping_list)
                ),
                jwt=st.session_state.jwt_val,
                filter_dict=filter_selection,
            )
        render_matches()

        add_social_share_buttons()


def get_info_from_endpoint(params, endpoint) -> dict:
    parameters = {}
    if params.secured:
        parameters['jwt'] = {'token': st.session_state.jwt_val['token']}
    response = call_flow(
        f'http://localhost:{NOWGATEWAY_BFF_PORT}/api/v1/info/{endpoint}',
        parameters,
        endpoint,
    )
    # return the first value of the response as it's a dict with one key
    return list(response.values())[0]


# hack to access cookies in streamlit: https://github.com/streamlit/streamlit/issues/5166#issuecomment-1259901215
# can also use https://gist.github.com/asehmi/85c293df44d32957fb31a7cd332c3398 but this one is shorter
def st_runtime():

    global _st_runtime

    if _st_runtime:
        return _st_runtime

    for obj in gc.get_objects():
        if type(obj) is Runtime:
            _st_runtime = obj
            return _st_runtime


_st_runtime = None
runtime = st_runtime()


def get_cookie(cookie_name):
    session_id = add_script_run_ctx().streamlit_script_run_ctx.session_id
    session_info = runtime._get_session_info(session_id)
    header = session_info.client.request.headers
    cookie_strings = [header_str for k, header_str in header.get_all() if k == 'Cookie']
    parsed_cookies = {k: v for c in cookie_strings for k, v in parse_cookie(c).items()}

    return parsed_cookies.get(cookie_name, '')


def render_auth_components(params):
    if params.secured:
        st_cookie = get_cookie(SSO_COOKIE)
        resp_jwt = requests.get(
            url=f'https://api.hubble.jina.ai/v2/rpc/user.identity.whoami',
            cookies={SSO_COOKIE: st_cookie},
        ).json()
        redirect_to = None
        if resp_jwt['code'] != 200:
            redirect_to = _do_login(params)
        else:
            st.session_state.login = False
            if not st.session_state.jwt_val:
                new_resp = {'token': st_cookie, 'user': resp_jwt['data']}
                st.session_state.jwt_val = new_resp
            if not st.session_state.avatar_val:
                st.session_state.avatar_val = resp_jwt['data']['avatarUrl']
            if not st.session_state.token_val:
                st.session_state.token_val = st_cookie

        if not st.session_state.jwt_val:
            redirect_to = _do_login(params)
        _, logout, avatar = st.columns([0.7, 0.12, 0.12])
        if not st.session_state.login:
            with avatar:
                if st.session_state.avatar_val:
                    st.image(st.session_state.avatar_val, width=30)
            with logout:
                st.button('Logout', on_click=_do_logout)
        return redirect_to
    else:
        return None


def _do_login(params):
    flow_namespace = os.environ.get("K8S_NAMESPACE_NAME", "").split('-')[1]
    jcloud_name = create_jcloud_name(params.flow_name) + '-' + flow_namespace
    host = f'https://{jcloud_name}-http.wolf.jina.ai/playground'
    redirect_uri = f'{host}/playground/'
    if 'top_k' in st.experimental_get_query_params():
        redirect_uri += f'?top_k={params.top_k}'

    redirect_uri = quote(redirect_uri)
    redirect_uri = (
        'https://api.hubble.jina.ai/v2/oidc/authorize?prompt=login&target_link_uri='
        + redirect_uri
    )
    st.session_state.login = True
    return redirect_uri


def _do_logout():
    headers = {
        'Content-Type': 'application/json; charset=utf-8',
        'Authorization': 'Token ' + st.session_state.token_val,
    }
    st.session_state.jwt_val = None
    st.session_state.avatar_val = None
    st.session_state.token_val = None
    st.session_state.login = True
    requests.post(
        'https://api.hubble.jina.ai/v2/rpc/user.session.dismiss',
        headers=headers,
    )


def load_example_queries(data):
    da_img = None
    da_txt = None
    if data in ds_set:
        try:
            da_img = load_data(S3_DEMO_PATH + data + f'.img10.bin')
            da_txt = load_data(S3_DEMO_PATH + data + f'.txt10.bin')
        except HTTPError as exc:
            print('Could not load samples for the demo dataset', exc)
    return da_img, da_txt


def setup_design():
    class UI:
        about_block = """
        ### About
        This is a meme search engine using [Jina's neural search framework](https://github.com/jina-ai/jina/).
        - [Live demo](https://examples.jina.ai/memes)
        - [Play with it in a notebook](https://colab.research.google.com/github/jina-ai/workshops/blob/main/memes/meme_search.ipynb) (t-only)
        - [Repo](https://github.com/alexcg1/jina-meme-search)
        - [Dataset](https://www.kaggle.com/abhishtagatya/imgflipscraped-memes-caption-dataset)
        """

        css = """
        <style>
            .reportview-container .main .block-container{{
                max-width: 1200px;
                padding-top: 2rem;
                padding-right: 2rem;
                padding-left: 2rem;
                padding-bottom: 2rem;
            }}
            .reportview-container .main {{
                color: "#111";
                background-color: "#eee";
            }}
        </style>
        """

    # Layout
    st.markdown(
        body=UI.css,
        unsafe_allow_html=True,
    )
    # design and create toggle button
    st.write(
        '<style>div.row-widget.stRadio > div{flex-direction:row;justify-content: center;} </style>',
        unsafe_allow_html=True,
    )
    st.write(
        '<style>div.st-bf{flex-direction:column;} div.st-ag{font-weight:bold;padding-right:50px;}</style>',
        unsafe_allow_html=True,
    )


def delete_semantic_scores():
    st.session_state['len_semantic_scores'] = 0
    st.session_state.semantic_scores = {}


def toggle_bm25_slider():
    if st.session_state.show_bm25_slider:
        st.session_state.show_bm25_slider = False
    else:
        st.session_state.show_bm25_slider = True


def get_encoder_options(q_field: str, id_field: str) -> List[str]:
    encoders_options = [
        encoder
        for encoder in st.session_state.index_fields_dict.keys()
        if id_field in st.session_state.index_fields_dict[encoder].keys()
    ]
    if q_field.startswith('image'):
        modality_models = [model['value'] for model in MODALITY_TO_MODELS[Image]]
    elif q_field.startswith('text'):
        modality_models = [model['value'] for model in MODALITY_TO_MODELS[Text]]
    else:
        raise ValueError(f'Unknown modality for field {q_field}')
    return list(set(encoders_options) & set(modality_models))


def customize_semantic_scores():
    input_modalities = [
        field['modality'] for field in list(st.session_state.query.values())
    ]
    add, delete, bm25 = st.columns([0.3, 0.3, 0.3])
    if add.button('Add semantic score', key='sem_score_button'):
        st.session_state['len_semantic_scores'] += 1
    if st.session_state.len_semantic_scores > 0:
        delete.button(
            label='Delete all semantic scores',
            key='delete',
            on_click=delete_semantic_scores,
        )
    if 'text' in input_modalities and any(
        field_mod == 'text'
        for field_mod in [
            st.session_state.index_fields_dict[encoder][field]
            for encoder in st.session_state.index_fields_dict.keys()
            for field in st.session_state.index_fields_dict[encoder].keys()
        ]
    ):
        bm25.button('Add bm25 score', key='bm25', on_click=toggle_bm25_slider)
        if st.session_state.show_bm25_slider:
            query_field_selectbox, bm25_slider = st.columns([0.5, 0.5])
            q_field = query_field_selectbox.selectbox(
                'Select query field for bm25 scoring',
                options=[
                    field
                    for field in st.session_state.query.keys()
                    if field.startswith('text')
                ],
            )
            bm25_weight = bm25_slider.slider(
                label='Adjust bm25 weight',
                min_value=0.0,
                max_value=1.0,
                value=0.5,
                key='weight_bm25',
            )
            st.session_state.semantic_scores['bm25'] = [
                q_field,
                'bm25_text',
                'bm25',
                bm25_weight,
            ]

    for i in range(st.session_state['len_semantic_scores']):
        query_field, index_field, encoder, weight = st.columns([0.25, 0.25, 0.25, 0.25])
        q_field = query_field.selectbox(
            label='query field',
            options=list(st.session_state.query.keys()),
            key='query_field_' + str(i),
        )
        id_field = index_field.selectbox(
            label='index field',
            options=list(st.session_state.field_names_to_dataclass_fields.keys()),
            key='index_field_' + str(i),
        )
        id_field = st.session_state.field_names_to_dataclass_fields[id_field]
        encoder_options = get_encoder_options(q_field, id_field)
        enc = encoder.selectbox(
            label='encoder',
            options=encoder_options,
            key='encoder_' + str(i),
        )
        w = weight.slider(
            label='weight',
            min_value=0.0,
            max_value=1.0,
            value=0.5,
            key='weight_' + str(i),
        )
        st.session_state.semantic_scores[f'{i}'] = (q_field, id_field, enc, w)


def render_mm_query(query, modality):
    if st.button("\+", key=f'add_{modality}_field'):  # noqa: W605
        st.session_state[f"len_{modality}_choices"] += 1
    if modality == 'text':
        for field_number in range(st.session_state[f"len_{modality}_choices"]):
            key = f'{modality}_{field_number}'
            query[key] = {
                'name': key,
                'value': st.text_input(
                    label=f'text #{field_number}',
                    key=key,
                    on_change=clear_match,
                    placeholder=f'Write your text query #{field_number}',
                ),
                'modality': 'text',
            }

    else:
        for field_number in range(st.session_state[f"len_{modality}_choices"]):
            key = f'{modality}_{field_number}'
            uploaded_image = st.file_uploader(
                label=f'image #{field_number}', key=key, on_change=clear_match
            )
            if uploaded_image:
                doc = convert_file_to_document(uploaded_image)
                query_doc = doc
                if query_doc.blob == b'':
                    if query_doc.tensor is not None:
                        query_doc.convert_image_tensor_to_blob()
                    elif (query_doc.uri is not None) and query_doc.uri != '':
                        query_doc.load_uri_to_blob(timeout=10)
                query[key] = {
                    'name': key,
                    'value': base64.b64encode(query_doc.blob).decode('utf-8'),
                    'modality': 'image',
                }
    if st.session_state[f"len_{modality}_choices"] >= 1:
        st.button(
            label=f'\-',  # noqa: W605
            key=f'remove_{modality}_field',
            on_click=decrement_inputs,
            kwargs=dict(modality=modality),
        )


def render_matches():
    # TODO function is too large. Split up.
    if st.session_state.matches and not st.session_state.error_msg:
        list_matches = [
            st.session_state.matches[i : i + 9]
            for i in range(0, len(st.session_state.matches), 9)
        ]

        # render the current page or the last page if filtered documents are less
        if list_matches:
            st.session_state.page_number = min(
                st.session_state.page_number, len(list_matches) - 1
            )
            st.header('Search results')
            # Results area
            c1, c2, c3 = st.columns(3)
            c4, c5, c6 = st.columns(3)
            c7, c8, c9 = st.columns(3)
            all_cs = [c1, c2, c3, c4, c5, c6, c7, c8, c9]
            for c, match in zip(all_cs, list_matches[st.session_state.page_number]):
                render_multi_modal_result(match, c)

        if len(list_matches) > 1:
            # disable prev button or not
            if st.session_state.page_number <= 0:
                st.session_state.disable_prev = True
            else:
                st.session_state.disable_prev = False

            # disable next button or not
            if st.session_state.page_number + 1 >= len(list_matches):
                st.session_state.disable_next = True
            else:
                st.session_state.disable_next = False

            prev, _, page, _, next = st.columns([1, 4, 2, 4, 1])
            page.write(f'Page {st.session_state.page_number + 1}/{len(list_matches)}')
            next.button(
                'Next', disabled=st.session_state.disable_next, on_click=increment_page
            )
            prev.button(
                'Previous',
                disabled=st.session_state.disable_prev,
                on_click=decrement_page,
            )

    if st.session_state.error_msg:
        with st.expander(
            'Received error response from the server. Expand this to see the full error message'
        ):
            st.text(st.session_state.error_msg)


def render_multi_modal_result(match, c):
    for chunk in match.chunks:
        render_graphic_result(chunk, c)
        render_text_result(chunk, c)
    if st.session_state.show_score_breakdown:
        render_score_breakdown(match, c)


# I'm not so happy about these two functions, let's refactor them
def render_graphic_result(match, c):
    try:
        match.mime_type = 'text-or-image-or-video'
        if match.blob:
            match.convert_blob_to_datauri()
        elif match.tensor is not None:
            match.convert_image_tensor_to_uri()
        c.image(match.uri)
    except:
        pass


def render_text_result(match, c):
    try:
        if match.text == '' and match.uri != '':
            match.load_uri_to_text(timeout=10)
        display_text = profanity.censor(match.text).replace('\n', ' ')
        body = f"<!DOCTYPE html><html><body>{display_text}</body></html>"
        c.markdown(
            body=body,
            unsafe_allow_html=True,
        )
    except:
        pass


def render_score_breakdown(match, c):
    display_scores_string = "<br>".join(
        sorted(
            [
                name + ": " + str(round(score.value, 3))
                for name, score in match.scores.items()
            ]
        )
    )
    body = f'<!DOCTYPE html><html><body><p style = "font-size:10px;">{display_scores_string}</body></html>'
    c.markdown(
        body=body,
        unsafe_allow_html=True,
    )


def add_social_share_buttons():
    # Adding social share buttons
    _, twitter, linkedin, facebook = st.columns([0.55, 0.12, 0.12, 0.12])
    for column, name in [
        (twitter, 'twitter'),
        (linkedin, 'linkedin'),
        (facebook, 'facebook'),
    ]:
        with column:
            components.html(BUTTONS[name])


def display_song(attach_to, song_doc: Document):
    genre = (
        song_doc.tags['genre_tags'].split(' ')[:3]
        if isinstance(song_doc.tags['genre_tags'], str)
        else song_doc.tags['genre_tags'][:3]
    )
    attach_to.markdown(
        body=f"<!DOCTYPE html><html><body>"
        f"<p style=\"font-size: 20px; font-weight: 700; margin-bottom: -8px\">{song_doc.tags['name']}</p>"
        f"<p style=\"margin-bottom: -5px\">{song_doc.tags['artist']}</p>"
        f"<p style=\"font-size: 10px\">{' | '.join(genre)}</p>"
        f"</body></html>",
        unsafe_allow_html=True,
    )
    if 'album_cover_image_url' in song_doc.tags:
        attach_to.markdown(
            body="<!DOCTYPE html><html><body>"
            f"<img/ src={song_doc.tags['album_cover_image_url']} style=\"width: 200px; height: 200px; padding: 10px 0 10px 0\">"
            "</body></html>",
            unsafe_allow_html=True,
        )
    attach_to.audio(io.BytesIO(song_doc.blob))


def increment_page():
    st.session_state.page_number += 1


def decrement_page():
    st.session_state.page_number -= 1


def decrement_inputs(modality):
    st.session_state[f"len_{modality}_choices"] -= 1


def clear_match():
    st.session_state.matches = (
        None  # TODO move this to when we choose a suggestion or search button
    )
    st.session_state.slider = 0.0
    st.session_state.snap = None
    st.session_state.error_msg = None
    st.session_state.page_number = 0


def clear_text():
    st.session_state.text_search_box = ''
    clear_match()


def load_data(data_path: str) -> DocumentArray:
    if data_path.startswith('http'):
        os.makedirs('data/tmp', exist_ok=True)
        url = data_path
        data_path = (
            f"data/tmp/{base64.b64encode(bytes(url, 'utf-8')).decode('utf-8')}.bin"
        )
        if not os.path.exists(data_path):
            with urlopen(url) as f:
                content = f.read()
            with open(data_path, 'wb') as f:
                f.write(content)

    try:
        da = DocumentArray.load_binary(data_path)
    except Exception:
        da = DocumentArray.load_binary(data_path, compress='gzip')
    return da


def get_login_button(url):
    return (
        f'<a href="{url}" target="_self" class="button">'
        + 'Login'
        + """<style>
                                    .button {
                                      margin-top: -50px;
                                      position: relative;
                                      overflow: hidden;
                                      -webkit-transition: background 400ms;
                                      transition: background 400ms;
                                      color: #fff;
                                      background-color: #90ee90;
                                      padding: 0.5em 0.5rem;
                                      font-family: 'Roboto', sans-serif;
                                      font-size: 1.0rem;
                                      outline: 0;
                                      border: 0;
                                      border-radius: 0.05rem;
                                      -webkit-box-shadow: 0 0 0.5rem rgba(0, 0, 0, 0.3);
                                      box-shadow: 0 0 0.5rem rgba(0, 0, 0, 0.3);
                                      cursor: pointer;
                                    }
                                    </style>
                                """
        + '</a>'
    )


def get_logout_button(url):
    return (
        f'<a href="{url}" target="_self" class="button">'
        + 'Logout'
        + """<style>
                                    .button {
                                      margin-top: -50px;
                                      position: relative;
                                      overflow: hidden;
                                      -webkit-transition: background 400ms;
                                      transition: background 400ms;
                                      color: #fff;
                                      background-color: #90ee90;
                                      padding: 0.5em 0.5rem;
                                      font-family: 'Roboto', sans-serif;
                                      font-size: 1.0rem;
                                      outline: 0;
                                      border: 0;
                                      border-radius: 0.05rem;
                                      -webkit-box-shadow: 0 0 0.5rem rgba(0, 0, 0, 0.3);
                                      box-shadow: 0 0 0.5rem rgba(0, 0, 0, 0.3);
                                      cursor: pointer;
                                    }
                                    </style>
                                """
        + '</a>'
    )


def setup_session_state():
    if 'matches' not in st.session_state:
        st.session_state.matches = None

    if 'im' not in st.session_state:
        st.session_state.im = None

    if 'snap' not in st.session_state:
        st.session_state.snap = None

    if 'search_count' not in st.session_state:
        st.session_state.search_count = 0

    if 'jwt_val' not in st.session_state:
        st.session_state.jwt_val = None

    if 'avatar_val' not in st.session_state:
        st.session_state.avatar_val = None

    if 'token_val' not in st.session_state:
        st.session_state.token_val = None

    if 'login' not in st.session_state:
        st.session_state.login = False

    if 'error_msg' not in st.session_state:
        st.session_state.error_msg = None

    if 'page_number' not in st.session_state:
        st.session_state.page_number = 0

    if 'disable_next' not in st.session_state:
        st.session_state.disable_next = True

    if 'disable_prev' not in st.session_state:
        st.session_state.disable_prev = True

    if 'filters' not in st.session_state:
        st.session_state.filters = 'notags'

    if 'filters_set' not in st.session_state:
        st.session_state.filters_set = False

    if "len_text_choices" not in st.session_state:
        st.session_state["len_text_choices"] = 1

    if "len_image_choices" not in st.session_state:
        st.session_state["len_image_choices"] = 1

    if "query" not in st.session_state:
        st.session_state['query'] = dict()

    if 'len_semantic_scores' not in st.session_state:
        st.session_state['len_semantic_scores'] = 0

    if 'index_fields_dict' not in st.session_state:
        st.session_state.index_fields_dict = {}

    if 'field_names_to_dataclass_fields' not in st.session_state:
        st.session_state.field_names_to_dataclass_fields = {}

    if 'encoder' not in st.session_state:
        st.session_state.encoder = 'clip'

    if 'semantic_scores' not in st.session_state:
        st.session_state.semantic_scores = {}

    if 'show_bm25_slider' not in st.session_state:
        st.session_state.show_bm25_slider = False

    if 'show_score_breakdown' not in st.session_state:
        st.session_state.show_score_breakdown = False


if __name__ == '__main__':
    # read user_input from user_input.json in the home directory
    with open(os.path.join(os.path.expanduser('~'), 'user_input.json'), 'r') as f:
        user_input_dict = json.load(f)

    user_input = UserInput()
    for attr_name, prev_value in user_input.__dict__.items():
        setattr(
            user_input,
            attr_name,
            user_input_dict.get(attr_name, prev_value),
        )

    deploy_streamlit(user_input=user_input)



import asyncio
import subprocess

from jcloud.flow import CloudFlow


def deploy_wolf(path: str):
    return CloudFlow(path=path).__enter__()


def terminate_wolf(flow_id: str):
    CloudFlow(flow_id=flow_id).__exit__()


def status_wolf(flow_id):
    loop = get_or_create_eventloop()
    return loop.run_until_complete(CloudFlow(flow_id=flow_id).status)


def get_or_create_eventloop():
    try:
        return asyncio.get_event_loop()
    except RuntimeError as ex:
        if "There is no current event loop in thread" in str(ex):
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            return asyncio.get_event_loop()


def list_all_wolf(status='Serving', namespace='nowapi'):
    flows = []
    loop = get_or_create_eventloop()
    jflows = loop.run_until_complete(CloudFlow().list_all(phase=status))['flows']
    # Transform the JCloud flow response to a much simpler list of dicts
    for flow in jflows:
        executor_name = list(flow['status']['endpoints'].keys())[0]
        flows.append(
            {'id': flow['id'], 'name': flow['status']['endpoints'][executor_name]}
        )
    # filter by namespace - if the namespace is contained in the flow name
    if namespace:
        return [f for f in flows if namespace in f['id']]
    return flows


def cmd(command, std_output=False, wait=True):
    if isinstance(command, str):
        command = command.split()
    if not std_output:
        process = subprocess.Popen(
            command, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
    else:
        process = subprocess.Popen(command)
    if wait:
        output, error = process.communicate()
        return output, error


def which(executable: str) -> bool:
    return bool(cmd('which ' + executable)[0])



"""
This module implements a command-line dialog with the user.
Its goal is to configure a UserInput object with users specifications.
Optionally, values can be passed from the command-line when jina-now is launched. In that case,
the dialog won't ask for the value.
"""
from __future__ import annotations, print_function, unicode_literals

import dataclasses
from copy import deepcopy
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, TypeVar, Union

from pydantic import BaseModel, StrictBool

from now.constants import DatasetTypes
from now.utils import docarray_typing_to_modality_string


class UserInput(BaseModel):
    app_instance: Optional['JinaNOWApp'] = None  # noqa: F821
    # data related
    flow_name: Optional[str] = None
    dataset_type: Optional[DatasetTypes] = None
    dataset_name: Optional[str] = None
    dataset_url: Optional[str] = None
    dataset_path: Optional[str] = None

    # AWS related
    aws_access_key_id: Optional[str] = None
    aws_secret_access_key: Optional[str] = None
    aws_region_name: Optional[str] = None

    # Fields
    index_fields: Optional[List] = []
    index_field_candidates_to_modalities: Optional[Dict[str, TypeVar]] = {}
    filter_fields: Optional[List] = []
    filter_field_candidates_to_modalities: Optional[Dict[str, str]] = {}
    field_names_to_dataclass_fields: Optional[Dict[str, str]] = {}
    model_choices: Optional[Dict[str, List]] = {}

    # ES related
    es_index_name: Optional[str] = None
    es_host_name: Optional[str] = None
    es_additional_args: Optional[Dict] = None

    # cluster related
    cluster: Optional[str] = None
    secured: Optional[StrictBool] = False
    jwt: Optional[Dict[str, str]] = None
    admin_name: Optional[str] = None
    admin_emails: Optional[List[str]] = None
    user_emails: Optional[List[str]] = None
    additional_user: Optional[StrictBool] = None
    api_key: Optional[str] = None

    class Config:
        arbitrary_types_allowed = True

    def to_safe_dict(self) -> Dict[str, Any]:
        user_input_dict = deepcopy(self.__dict__)
        user_input_dict.pop('app_instance', None)
        user_input_dict['index_field_candidates_to_modalities'] = {
            field: docarray_typing_to_modality_string(modality)
            for field, modality in user_input_dict[
                'index_field_candidates_to_modalities'
            ].items()
        }
        return user_input_dict


@dataclasses.dataclass
class DialogOptions:
    name: str
    prompt_message: str
    prompt_type: str
    choices: Union[
        List[Dict[str, Union[str, bool]]],
        Callable[[Any], List[Dict[str, str]]],
    ] = None
    is_terminal_command: StrictBool = (
        False  # set when this dialog is required as a cli param
    )
    argparse_kwargs: Dict[str, Any] = dataclasses.field(default_factory=dict)
    description: str = None  # Description to show on terminal when used as a cli param
    depends_on: Optional['DialogOptions', StrictBool] = None
    default: Optional[str] = None
    conditional_check: Callable[[Any], bool] = None
    post_func: Callable[[Any], None] = None
    dynamic_func: Callable[[Any], List[DialogOptions]] = None


if TYPE_CHECKING:
    from now.app.base.app import JinaNOWApp

    UserInput.update_forward_refs()



from __future__ import annotations, print_function, unicode_literals

import base64
import json
import os
import tempfile
from concurrent.futures import ThreadPoolExecutor

import boto3
from docarray import Document, DocumentArray

from now.utils import flatten_dict


def maybe_download_from_s3(
    docs: DocumentArray, tmpdir: tempfile.TemporaryDirectory, user_input, max_workers
):
    """Downloads file to local temporary dictionary, saves S3 URI to `tags['uri']` and modifies `uri` attribute of
    document to local path in-place.

    :param doc: document containing URI pointing to the location on S3 bucket
    :param tmpdir: temporary directory in which files will be saved
    :param user_input: User iput which contain aws credentials
    :param max_workers: number of threads to create in the threadpool executor to make execution faster
    """

    flat_docs = docs['@c']
    filtered_docs = [c for c in flat_docs if c.uri.startswith('s3://')]

    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        futures = []
        for c in filtered_docs:
            f = executor.submit(
                convert_fn,
                c,
                tmpdir,
                user_input.aws_access_key_id,
                user_input.aws_secret_access_key,
                user_input.aws_region_name,
            )
            futures.append(f)
        for d in docs:
            f = executor.submit(
                update_tags,
                d,
                user_input.aws_access_key_id,
                user_input.aws_secret_access_key,
                user_input.aws_region_name,
            )
            futures.append(f)
        for f in futures:
            f.result()


def convert_fn(
    d: Document, tmpdir, aws_access_key_id, aws_secret_access_key, aws_region_name
) -> Document:
    """Downloads files and tags from S3 bucket and updates the content uri and the tags uri to the local path"""

    bucket = get_bucket(
        uri=d.uri,
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        region_name=aws_region_name,
    )
    d.tags['uri'] = d.uri

    d.uri = download_from_bucket(tmpdir, d.uri, bucket)
    if d.uri.endswith('.json'):
        d.load_uri_to_text()
        json_dict = json.loads(d.text)
        field_name = d._metadata['field_name']
        field_value = get_dict_value_for_flattened_key(
            json_dict, field_name.split('__')
        )
        d.text = field_value
        d.uri = ''
    return d


def get_bucket(uri, aws_access_key_id, aws_secret_access_key, region_name):
    session = boto3.session.Session(
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        region_name=region_name,
    )
    bucket = session.resource('s3').Bucket(uri.split('/')[2])
    return bucket


def download_from_bucket(tmpdir, uri, bucket):
    path_s3 = '/'.join(uri.split('/')[3:])
    path_local = get_local_path(tmpdir, path_s3)
    bucket.download_file(
        path_s3,
        path_local,
    )
    return path_local


def get_dict_value_for_flattened_key(d, keys):
    if len(keys) == 0:
        return d
    else:
        return get_dict_value_for_flattened_key(d[keys[0]], keys[1:])


def get_local_path(tmpdir, path_s3):
    # todo check if this method of creatign the path is creating too much overhead
    # also, the number of files is growing and will never be cleaned up
    return os.path.join(
        str(tmpdir),
        base64.b64encode(bytes(path_s3, "utf-8")).decode("utf-8")
        + f'.{path_s3.split(".")[-1] if "." in path_s3 else ""}',  # preserve file ending
    )


def update_tags(d, aws_access_key_id, aws_secret_access_key, region_name):
    if '_s3_uri_for_tags' in d._metadata:
        bucket = get_bucket(
            uri=d._metadata['_s3_uri_for_tags'],
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            region_name=region_name,
        )
        with tempfile.TemporaryDirectory() as tmpdir:
            local_file = download_from_bucket(
                tmpdir, d._metadata['_s3_uri_for_tags'], bucket
            )

            with open(local_file, 'r') as file:
                data = json.load(file)

        d.tags.update(flatten_dict(data))



from __future__ import annotations, print_function, unicode_literals

import os
import shutil
import signal
import sys
from collections.abc import MutableMapping
from dataclasses import dataclass
from inspect import stack
from typing import Any, Dict, List, Optional, TypeVar, Union

import boto3
import docarray
import hubble
import yaml
from jina.jaml import JAML
from pyfiglet import Figlet

from now.thirdparty.PyInquirer.prompt import prompt


def download_file(path, r_raw):
    with path.open("wb") as f:
        shutil.copyfileobj(r_raw, f)


def download(url, filename):
    import functools
    import pathlib

    import requests
    from tqdm.auto import tqdm

    r = requests.get(url, stream=True, allow_redirects=True)
    if r.status_code != 200:
        r.raise_for_status()  # Will only raise for 4xx codes, so...
        raise RuntimeError(f"Request to {url} returned status code {r.status_code}")
    file_size = int(r.headers.get('Content-Length', 0))

    path = pathlib.Path(filename).expanduser().resolve()
    path.parent.mkdir(parents=True, exist_ok=True)

    desc = "(Unknown total file size)" if file_size == 0 else ""
    r.raw.read = functools.partial(
        r.raw.read, decode_content=True
    )  # Decompress if needed

    if any(map(lambda x: x in os.environ, ['NOW_CI_RUN', 'NOW_EXAMPLES'])):
        download_file(path, r.raw)
    else:
        with tqdm.wrapattr(r.raw, "read", total=file_size, desc=desc) as r_raw:
            download_file(path, r_raw)

    return path


def my_handler(signum, frame, spinner):
    with spinner.hidden():
        sys.stdout.write("Program terminated!\n")
    spinner.stop()
    exit(0)


class BetterEnum:
    def __iter__(self):
        return [getattr(self, x) for x in dir(self) if ('__' not in x)].__iter__()


def to_camel_case(text):
    """Convert text to camel case in great coding style"""
    s = text.replace("_", " ")
    s = s.split()
    return ''.join(i.capitalize() for i in s)


sigmap = {signal.SIGINT: my_handler, signal.SIGTERM: my_handler}


class EnvironmentVariables:
    def __init__(self, envs: Dict):
        self._env_keys_added: Dict = envs

    def __enter__(self):
        for key, val in self._env_keys_added.items():
            os.environ[key] = str(val)

    def __exit__(self, *args, **kwargs):
        for key in self._env_keys_added.keys():
            os.unsetenv(key)


def write_env_file(env_file, config):
    config_string = '\n'.join([f'{key}={value}' for key, value in config.items()])
    with open(env_file, 'w+') as fp:
        fp.write(config_string)


def write_flow_file(flow_yaml_content, new_yaml_file_path):
    with open(new_yaml_file_path, 'w') as f:
        JAML.dump(
            flow_yaml_content,
            f,
            indent=2,
            allow_unicode=True,
            Dumper=Dumper,
        )


@hubble.login_required
def jina_auth_login():
    pass


def get_info_hubble(user_input):
    client = hubble.Client(max_retries=None, jsonify=True)
    response = client.get_user_info()
    user_input.admin_emails = (
        [response['data']['email']] if 'email' in response['data'] else []
    )
    if not user_input.admin_emails:
        print(
            'Your hubble account is not verified. Please verify your account to deploy your flow as admin.'
        )
    user_input.jwt = {'token': client.token}
    user_input.admin_name = response['data']['name']
    return response['data'], client.token


def print_headline():
    f = Figlet(font='slant')
    print('Welcome to:')
    print(f.renderText('Jina NOW'))
    print('Get your search use case up and running - end to end.\n')
    print(
        'You can choose between image and text search. \nJina NOW trains a model, pushes it to Jina AI Cloud '
        'and deploys a Flow and playground app in the cloud or locally. \nCheck out our demos or bring '
        'your own data.\n'
    )
    print('Visit docs.jina.ai to learn more about our framework')
    print(
        '💡 Make sure you give enough memory to your Docker daemon. '
        '5GB - 8GB should be okay.'
    )
    print()


def maybe_prompt_user(questions, attribute, **kwargs):
    """
    Checks the `kwargs` for the `attribute` name. If present, the value is returned directly.
    If not, the user is prompted via the cmd-line using the `questions` argument.

    :param questions: A dictionary that is passed to `PyInquirer.prompt`
        See docs: https://github.com/CITGuru/PyInquirer#documentation
    :param attribute: Name of the value to get. Make sure this matches the name in `kwargs`

    :return: A single value of either from `kwargs` or the user cli input.
    """
    if kwargs and attribute in kwargs:
        return kwargs[attribute]
    else:
        answer = prompt(questions)
        return answer[attribute]


def prompt_value(
    name: str,
    prompt_message: str,
    prompt_type: str = 'input',
    choices: Optional[List[Union[Dict, str]]] = None,
    **kwargs: Dict,
):
    qs = {'name': name, 'type': prompt_type, 'message': prompt_message}

    if choices is not None:
        qs['choices'] = choices
    return maybe_prompt_user(qs, name, **kwargs)


def debug(msg: Any):
    """
    Prints a message along with the details of the caller, ie filename and line number.
    """
    msg = str(msg) if msg is not None else ''
    frameinfo = stack()[1]
    print(f"{frameinfo.filename}:{frameinfo.lineno}: {msg}")


def flatten_dict(d, parent_key='', sep='__'):
    """
    This function converts a nested dictionary into a dictionary of attirbutes using '__' as a separator.
    Example:
        {'a': {'b': {'c': 1, 'd': 2}}} -> {'a__b__c': 1, 'a__b__c': 2}
    """
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, MutableMapping):
            items.extend(flatten_dict(v, new_key, sep=sep).items())
        else:
            # TODO for now, we just have string values, str(v) should be removed once we support numeric values
            items.append((new_key, str(v)))
    return dict(items)


def get_flow_id(host):
    return host[len('https://') : -len('-http.wolf.jina.ai')]


class Dumper(yaml.Dumper):
    def increase_indent(self, flow=False, *args, **kwargs):
        return super().increase_indent(flow=flow, indentless=False)


def docarray_typing_to_modality_string(T: TypeVar) -> str:
    """E.g. docarray.typing.Image -> image"""
    return T.__name__.lower()


def modality_string_to_docarray_typing(s: str) -> TypeVar:
    """E.g. image -> docarray.typing.Image"""
    return getattr(docarray.typing, s.capitalize())


@dataclass
class AWSProfile:
    aws_access_key_id: str
    aws_secret_access_key: str
    region: str


def get_aws_profile():
    session = boto3.Session()
    credentials = session.get_credentials()
    aws_profile = (
        AWSProfile(credentials.access_key, credentials.secret_key, session.region_name)
        if credentials
        else AWSProfile(None, None, session.region_name)
    )
    return aws_profile


def hide_string_chars(s):
    return ''.join(['*' for _ in range(len(s) - 5)]) + s[len(s) - 4 :] if s else None


def get_chunk_by_field_name(doc, field_name):
    """
    Gets a specific chunk by field name, using its position instead of getting the attribute directly.
    This solves the getattr problem when there are conflicting attributes with Document.

    :param doc: Document to get the chunk from.
    :param field_name: Field needed to extract the position.

    :return: Specific chunk by field.
    """
    try:
        field_position = int(
            doc._metadata['multi_modal_schema'][field_name]['position']
        )
        return doc.chunks[field_position]
    except Exception as e:
        print(f'An error occurred: {e}')


# Add a custom retry exception
class RetryException(Exception):
    pass


class DemoAvailableException(Exception):
    pass



"""
This module implements a command-line dialog with the user.
Its goal is to configure a UserInput object with users specifications.
Optionally, values can be passed from the command-line when jina-now is launched. In that case,
the dialog won't ask for the value.
"""
from __future__ import annotations, print_function, unicode_literals

import inspect
import pathlib

import now.utils
from now.common import options
from now.common.options import construct_app
from now.constants import MODALITY_TO_MODELS, Apps, DialogStatus
from now.now_dataclasses import DialogOptions, UserInput
from now.utils import DemoAvailableException, RetryException

cur_dir = pathlib.Path(__file__).parent.resolve()


def configure_user_input(**kwargs) -> UserInput:
    user_input = UserInput()
    now.utils.print_headline()
    # Create the search app.
    # TODO: refactor this when more apps are added
    user_input.app_instance = construct_app(Apps.SEARCH_APP)
    # Ask the options
    for option in options.base_options + user_input.app_instance.options:
        if configure_option(option, user_input, **kwargs) == DialogStatus.BREAK:
            break

    return user_input


def configure_option(
    option: DialogOptions, user_input: UserInput, **kwargs
) -> DialogStatus:
    # Check if it is dynamic. If it is then spawn multiple dialogs for each option and return with continue
    if option.dynamic_func:
        if option.name in kwargs:
            # Expand dynamic options from parent option, expect a dict (supports only model_selection)
            for user_selection in kwargs[option.name].split(","):
                if ":" in user_selection:
                    option_name, option_values = user_selection.split(":")
                    kwargs[f"{option_name}_model"] = []
                    if (
                        not option_name
                        in user_input.index_field_candidates_to_modalities
                    ):
                        raise ValueError(
                            f"Error with --{option.name}: `{option_name}` is not an index field."
                        )
                    for option_value in option_values.split("+"):
                        model_selection = [
                            model
                            for model in MODALITY_TO_MODELS[
                                user_input.index_field_candidates_to_modalities[
                                    option_name
                                ]
                            ]
                            if model["name"] == option_value
                        ]
                        if model_selection:
                            kwargs[f"{option_name}_model"].append(
                                model_selection[0]["value"]
                            )
                        else:
                            model_choices = [
                                model["name"]
                                for model in MODALITY_TO_MODELS[
                                    user_input.index_field_candidates_to_modalities[
                                        option_name
                                    ]
                                ]
                            ]
                            raise ValueError(
                                f"Error with --{option.name}: `{option_value}` is not available. "
                                f"for index field `{option_name}`. Choices are: {','. join(model_choices)}."
                            )

        for result in option.dynamic_func(user_input):
            configure_option(result, user_input, **kwargs)
        return DialogStatus.CONTINUE

    # Check if it is dependent on some other dialog options
    if option.depends_on and not option.conditional_check(user_input):
        return DialogStatus.SKIP

    # Populate choices if needed
    if option.choices and inspect.isfunction(option.choices):
        option.choices = option.choices(user_input, **kwargs)

    while True:
        val = now.utils.prompt_value(
            **option.__dict__,
            **kwargs,
        )

        if val:
            kwargs[option.name] = val
            if hasattr(user_input, option.name):
                setattr(user_input, option.name, val)

        try:
            # If there is any post function then invoke that
            if inspect.isfunction(option.post_func):
                option.post_func(user_input, **kwargs)
        except RetryException as e:
            print(e)
            continue
        except DemoAvailableException:
            return DialogStatus.BREAK
        break

    return DialogStatus.CONTINUE



import logging
from typing import Dict, Optional, Type, Union

from docarray import Document, DocumentArray

from now.data_loading.elasticsearch.connector import ElasticsearchConnector
from now.now_dataclasses import UserInput

logging.getLogger("PIL.Image").setLevel(logging.CRITICAL + 1)

ID_TAG = 'id'
FIELD_TAG = 'field_name'
EXTRACTION_TYPE_TAG = 'extraction_type'


class ElasticsearchExtractor:
    def __init__(
        self,
        query: Dict,
        index: str,
        user_input: UserInput,
        connection_str: str,
        data_class: Type = None,
        connection_args: Optional[Dict] = None,
    ):
        """
        For extracting documents from Elasticsearch into a `docarray.DocumentArray`
        dataset, this class implements an iterator which yields `docarray.Document`
        objects. To specify the data for extraction, one needs to provide an
        es query together with the index name and parameters to connect to
        the Elasticsearch instance.
        :param query: Elasticsearch query in the form of a JSON string
        :param index: Name of the ES index containing the documents to be extracted
        :param connection_str: A connection string for the ES instance. Usually, it
            includes url, port, username, password, etc. Typically, it has the form:
            'https://{user_name}:{password}@{host}:{port}'
        :param connection_args: Dictionary with additional connection arguments,
            e.g., information about certificates
        """
        self._es_connector = ElasticsearchConnector(
            connection_str=connection_str,
            connection_args=(connection_args if connection_args else {}),
        )
        self._query = query
        self._index = index
        self._user_input = user_input
        self._data_class = data_class
        self._document_cache = []
        self._query_result = self._es_connector.get_documents_by_query(
            self._query, self._index
        )

    def extract(self) -> DocumentArray:
        return DocumentArray([doc for doc in self._extract_documents()])

    def _extract_documents(self):
        try:
            next_doc = self._get_next_document()
            while next_doc:
                yield next_doc
                next_doc = self._get_next_document()
        except StopIteration:
            self._es_connector.close()
            return

    def _get_next_document(self) -> Union[Document, None]:
        """
        Returns the next document from the Elasticsearch database.
        In order to retrieve further documents, Elasticsearch documents are retrieved
        in pages of multiple documents. After retrieving a page, its contained
        documents are stored in a document cache. If documents are left in the cache
        this function returns one of those documents. Otherwise, the next page is
        queried. If there is no page left, None will be returned.
        :return: extracted document
        """
        if len(self._document_cache) == 0:
            self._document_cache = next(self._query_result)
            if len(self._document_cache) == 0:
                return None
        return self._construct_document(self._document_cache.pop())

    def _construct_document(self, es_document: Dict) -> Document:
        """
        Constructs a `docarray.Document` object from an Elasticsearch document.

        :param es_document: Elasticsearch document

        :return: `docarray.Document` object

        Creates a document using the dataclass specified in the user input.
        """
        kwargs, tags = {}, {}
        for field_name, field_value in es_document.items():
            if field_name in self._user_input.index_fields:
                kwargs[
                    self._user_input.field_names_to_dataclass_fields[field_name]
                ] = field_value
            else:
                tags[field_name] = field_value
        doc = Document(self._data_class(**kwargs))
        doc.tags = tags
        return doc



from typing import Dict, List

from pydantic import BaseModel, Field

from now.executor.gateway.bff.app.v1.models.shared import BaseRequestModel


class TagsResponseModel(BaseModel):
    tags: Dict[str, List] = Field(
        default={},
        description='Get all tags and their possible values in the index',
        example={
            'tags__price': [1, 2, 3],
            'tags__color': [
                'red',
                'blue',
                'green',
            ],
        },
    )


class CountRequestModel(BaseRequestModel):
    limit: int = Field(default=10, description='Maximum count', example=10)


class CountResponseModel(BaseModel):
    number_of_docs: int = Field(
        default=0,
        description='Get the number of documents in the index',
        example=100,
    )


class FieldNamesToDataclassFieldsResponseModel(BaseModel):
    field_names_to_dataclass_fields: Dict[str, str] = Field(
        default={},
        description='Dictionary which maps dataclass fields to their field names',
        example={'title': 'text_0', 'image': 'image_0'},
    )


class EncoderToDataclassFieldsModsResponseModel(BaseModel):
    encoder_to_dataclass_fields_mods: Dict[str, Dict[str, str]] = Field(
        default={},
        description='Dictionary which maps encoder names to the dataclass fields they encode and their modality',
        example={
            'encoderclip': {
                'text_0': 'text',
                'image_0': 'image',
            },
            'encodersbert': {
                'image_0': 'text',
            },
        },
    )



# -*- coding: utf-8 -*-
"""
`password` type question
"""
from . import input

# use std prompt-toolkit control


def question(message, **kwargs):
    kwargs['is_password'] = True
    return input.question(message, **kwargs)



import base64
from typing import List

from docarray import Document
from fastapi import APIRouter, Body

from now.data_loading.create_dataclass import create_dataclass
from now.executor.gateway.bff.app.v1.models.search import (
    SearchRequestModel,
    SearchResponseModel,
    SuggestionRequestModel,
)
from now.executor.gateway.bff.app.v1.routers.helper import (
    field_dict_to_mm_doc,
    jina_client_post,
)
from now.utils import get_chunk_by_field_name, modality_string_to_docarray_typing

search_examples = {
    'working_text': {
        'summary': 'A working example: search with text',
        'description': 'A working example which can be tried out. Search with text on the best artworks dataset.',
        'value': {
            'host': 'grpcs://now-example-best-artworks.dev.jina.ai',
            'port': 443,
            'limit': 10,
            'query': [
                {
                    'name': 'query_text_0',
                    'modality': 'text',
                    'value': 'cute cats',
                }
            ],
            'create_temp_link': False,
            'get_score_breakdown': True,
        },
    },
    'working_text_image': {
        'summary': 'A working example: search with text and image',
        'description': 'A working example which can be tried out. Search with text and image on the best artworks dataset.',
        'value': {
            'host': 'grpcs://now-example-best-artworks.dev.jina.ai',
            'port': 443,
            'limit': 10,
            'query': [
                {
                    'name': 'query_text_0',
                    'modality': 'text',
                    'value': 'cute cats',
                },
                {
                    'name': 'query_image_0',
                    'modality': 'image',
                    'value': 'https://images.pexels.com/photos/45201/kitty-cat-kitten-pet-45201.jpeg',
                },
            ],
            'create_temp_link': False,
            'get_score_breakdown': True,
        },
    },
    'dummy': {
        'summary': 'A dummy example',
        'description': 'A dummy example,  do not run. For parameter reference only.',
        'value': {
            'host': 'localhost',
            'port': 31080,
            'limit': 10,
            'filters': {
                'tags__color': {'$eq': 'blue'},
                'tags__price': {'$lte': 100, '$gte': 50},
            },
            'query': [
                {
                    'name': 'query_text_0',
                    'modality': 'text',
                    'value': 'cute cats',
                }
            ],
            'create_temp_link': False,
            'semantic_scores': [('query_text_0', 'title', 'encoderclip', 1.0)],
        },
    },
}

suggestion_examples = {
    'working_text': {
        'summary': 'A working example: get suggestions for a text query',
        'description': 'A working example which can be tried out. Get autocomplete suggestions for a text query.',
        'value': {
            'host': 'grpcs://now-example-best-artworks.dev.jina.ai',
            'port': 443,
            'text': 'cute ca',
        },
    },
    'dummy': {
        'summary': 'A dummy example',
        'description': 'A dummy example,  do not run. For parameter reference only.',
        'value': {
            'host': 'localhost',
            'port': 31080,
            'jwt': {'token': '<your token>'},
            'api_key': '<your api key>',
            'text': 'cute cats',
        },
    },
}

router = APIRouter()


@router.post(
    '/search',
    response_model=List[SearchResponseModel],
    summary='Search data via query',
)
async def search(
    data: SearchRequestModel = Body(examples=search_examples),
):
    fields_modalities_mapping = {}
    fields_values_mapping = {}

    if len(data.query) == 0:
        raise ValueError('Query cannot be empty')

    for field in data.query:
        fields_modalities_mapping[field['name']] = modality_string_to_docarray_typing(
            field['modality']
        )
        fields_values_mapping[field['name']] = field['value']
    data_class, field_names_to_dataclass_fields = create_dataclass(
        fields=list(fields_modalities_mapping.keys()),
        fields_modalities=fields_modalities_mapping,
    )
    query_doc = field_dict_to_mm_doc(
        fields_values_mapping,
        data_class=data_class,
        modalities_dict=fields_modalities_mapping,
        field_names_to_dataclass_fields=field_names_to_dataclass_fields,
    )
    query_filter = {}
    for key, value in data.filters.items():
        key = 'tags__' + key if not key.startswith('tags__') else key
        query_filter[key] = {'$eq': value}

    docs = await jina_client_post(
        endpoint='/search',
        docs=query_doc,
        parameters={
            'limit': data.limit,
            'filter': query_filter,
            'create_temp_link': data.create_temp_link,
            'semantic_scores': [
                list(semantic_score) for semantic_score in data.semantic_scores
            ],
            'get_score_breakdown': data.get_score_breakdown,
        },
        request_model=data,
    )
    matches = []
    for doc in docs[0].matches:
        # todo: use multimodal doc in the future!
        scores = {}
        for score_name, named_score in doc.scores.items():
            scores[score_name] = named_score.to_dict()
        # since multimodal doc is not supported, we take the first chunk
        if doc.chunks:
            field_names_and_chunks = [
                [field_name, get_chunk_by_field_name(doc, field_name)]
                for field_name in doc._metadata['multi_modal_schema'].keys()
            ]
        else:
            # TODO remove else path. It is only used to support the inmemory indexer since that one is operating on chunks while elastic responds with root documents
            field_names_and_chunks = [['result_field', doc]]
        results = {}
        for field_name, chunk in field_names_and_chunks:
            if chunk.blob:
                result = {'blob': base64.b64encode(chunk.blob).decode('utf-8')}
            elif chunk.text:
                result = {'text': chunk.text}
            elif chunk.uri:
                # in case we have content and uri, the content is preferred
                result = {'uri': chunk.uri}
            else:
                # We should not raise exception else it breaks the playground if a single chunk has no content
                # irrespective of what other chunks hold. We should just log it and move on.
                print('Result without content', doc.id, doc.tags)
                result = {'text': ''}
            results[field_name] = result
        match = SearchResponseModel(
            id=doc.id,
            scores=scores,
            tags=doc.tags,
            fields=results,
        )
        matches.append(match)
    return matches


@router.post(
    '/suggestion',
    summary='Get auto complete suggestion for query',
)
async def suggestion(data: SuggestionRequestModel = Body(examples=suggestion_examples)):
    suggest_doc = Document(text=data.text)
    docs = await jina_client_post(
        endpoint='/suggestion',
        docs=suggest_doc,
        request_model=data,
        target_executor=r'\Aautocomplete_executor\Z',
    )
    return docs.to_dict()



import os
import pathlib
import sys
import warnings
from argparse import Namespace

from now import __version__
from now import __version__ as version
from now.constants import SURVEY_LINK
from now.run_all_k8s import fetch_logs_now, start_now, stop_now

warnings.filterwarnings("ignore")

cur_dir = pathlib.Path(__file__).parents[1].resolve()

os.environ['JINA_CHECK_VERSION'] = 'False'
os.environ['JCLOUD_LOGLEVEL'] = 'ERROR'


def _get_run_args():
    from now.cli.parser import get_main_parser

    parser = get_main_parser()
    if len(sys.argv) == 1:
        parser.print_help()
        exit()
    args, unknown = parser.parse_known_args()

    # clean up the args with None values
    args = {k: v for k, v in vars(args).items() if v is not None}
    # Convert args back to Namespace
    args = Namespace(**args)

    if unknown:
        raise Exception('unknown args: ', unknown)

    return args


def get_task(kwargs):
    for x in ['cli', 'now']:
        if x in kwargs:
            return kwargs[x]
    raise Exception('kwargs do not contain a task')


def cli(args=None):
    """The main entrypoint of the CLI"""
    os.environ['JINA_LOG_LEVEL'] = 'CRITICAL'
    print_version_line()
    kwargs = parse_args(args)
    task = get_task(kwargs)
    if '--version' in sys.argv[1:]:
        print(__version__)
        exit(0)
    if task == 'start':
        return start_now(**kwargs)
    elif task == 'stop':
        return stop_now(**kwargs)
    elif task == 'logs':
        return fetch_logs_now(**kwargs)
    elif task == 'survey':
        import webbrowser

        webbrowser.open(SURVEY_LINK, new=0, autoraise=True)
    else:
        raise Exception(f'unknown task, {task}')


def parse_args(args):
    if not args:
        args = _get_run_args()
    args = vars(args)  # Make it a dict from Namespace
    return args


def print_version_line():
    if len(sys.argv) != 1 and not ('-h' in sys.argv[1:] or '--help' in sys.argv[1:]):
        print(f'Initialising Jina NOW v{version} ...')


if __name__ == '__main__':
    cli()



from typing import Any, Dict, Optional

from pydantic import BaseModel, Field, root_validator

from now.utils import get_chunk_by_field_name


class BaseRequestModel(BaseModel):
    jwt: Optional[Dict[str, Any]] = Field(
        default=None,
        description="Dictionary with key 'token' which maps to Jina Cloud token value."
        " To be passed when the flow is secure",
        example={'token': '<your jina cloud token>'},
    )
    api_key: Optional[str] = Field(
        default=None,
        description='Used to authenticate machines',
        example='<your api key>',
    )

    class Config:
        allow_mutation = False
        case_sensitive = False
        arbitrary_types_allowed = True


class ModalityModel(BaseModel):
    uri: Optional[str] = Field(
        default=None,
        description='URI of the file or data URI',
        example='https://example.com/image.jpg',
    )
    text: Optional[str] = Field(default=None, description='Text', example='cute cats')
    blob: Optional[str] = Field(
        default=None, description='Base64 encoded `utf-8` str format', example='xxx'
    )

    @root_validator(pre=True)
    def validate_only_one_exists(cls, values):
        # Get the names of all fields that are set (i.e. have a non-None value)
        set_fields = [name for name, value in values.items() if value is not None]
        if len(set_fields) != 1:
            raise ValueError(f"Only one of {set_fields} can be set.")
        return values

    @property
    def content(self):
        for field_name in self.__fields_set__:
            field_value = get_chunk_by_field_name(self, field_name)
            if field_value:
                return field_value



import cowsay
import requests
from rich import box
from rich.console import Console
from rich.panel import Panel
from rich.table import Column, Table

from now import run_backend
from now.constants import DEMO_NS, FLOW_STATUS
from now.deployment.deployment import cmd, list_all_wolf, status_wolf, terminate_wolf
from now.dialog import configure_user_input
from now.utils import maybe_prompt_user


def stop_now(**kwargs):
    choices = []
    # Add all remote Flows that exists with the namespace `nowapi`
    alive_flows = list_all_wolf(status=FLOW_STATUS)
    for flow_details in alive_flows:
        choices.append(flow_details['name'])
    if len(choices) == 0:
        cowsay.cow('nothing to stop')
        return
    else:
        questions = [
            {
                'type': 'list',
                'name': 'cluster',
                'message': 'Which cluster do you want to delete?',
                'choices': choices,
            }
        ]
        cluster = maybe_prompt_user(questions, 'cluster', **kwargs)

    flow = [x for x in alive_flows if x['name'] == cluster][0]
    flow_id = flow['id']
    _result = status_wolf(flow_id)
    if _result is None:
        print(f'❎ Flow not found in JCloud. Likely, it has been deleted already')
    if _result is not None and _result['status']['phase'] == FLOW_STATUS:
        terminate_wolf(flow_id)
        from hubble import Client

        cookies = {'st': Client().token}
        requests.delete(
            f'https://storefrontapi.nowrun.jina.ai/api/v1/schedule_sync/{flow_id}',
            cookies=cookies,
        )
    cowsay.cow(f'remote Flow `{cluster}` removed')


def start_now(**kwargs):
    user_input = configure_user_input(**kwargs)
    app_instance = user_input.app_instance
    # Only if the deployment is remote and the demo examples is available for the selected app
    # Should not be triggered for CI tests
    if app_instance.is_demo_available(user_input):
        gateway_host_grpc = f'grpcs://{DEMO_NS.format(user_input.dataset_name.split("/")[-1])}.dev.jina.ai'
    else:
        (
            gateway_port,
            gateway_host_grpc,
        ) = run_backend.run(app_instance, user_input, **kwargs)
    gateway_host_http = gateway_host_grpc.replace('grpc', 'http')
    bff_url = f'{gateway_host_http}/api/v1/search-app/docs'
    playground_url = f'{gateway_host_http}/playground'

    print()
    my_table = Table(
        'Attribute',
        Column(header="Value", overflow="fold"),
        show_header=False,
        box=box.SIMPLE,
        highlight=True,
    )
    my_table.add_row('Host (HTTPS)', gateway_host_http)
    my_table.add_row('Host (GRPCS)', gateway_host_grpc)
    my_table.add_row('API docs', bff_url)
    if user_input.secured and user_input.api_key:
        my_table.add_row('API Key', user_input.api_key)
    my_table.add_row('Playground', playground_url)
    console = Console()
    console.print(
        Panel(
            my_table,
            title=f':tada: Search app is NOW ready!',
            expand=False,
        )
    )
    return {
        'bff': bff_url,
        'playground': playground_url,
        'host_http': gateway_host_http,
        'host_grpc': gateway_host_grpc,
        'secured': user_input.secured,
    }


def fetch_logs_now(**kwargs):
    choices = []
    # Add all remote Flows that exists with the namespace `nowapi`
    alive_flows = list_all_wolf(status=FLOW_STATUS)
    for flow_details in alive_flows:
        choices.append(flow_details['name'])
    if len(choices) == 0:
        cowsay.cow('nothing to log')
        return
    else:
        questions = [
            {
                'type': 'list',
                'name': 'cluster',
                'message': 'Which cluster do you want to check logs for?',
                'choices': choices,
            }
        ]
        cluster = maybe_prompt_user(questions, 'cluster', **kwargs)

    flow = [x for x in alive_flows if x['name'] == cluster][0]
    flow_id = flow['id']
    _result = status_wolf(flow_id)
    if _result is None:
        print(f'❎ Flow not found in JCloud. Likely, it has been deleted already')

    if _result is not None and _result['status']['phase'] == FLOW_STATUS:
        namespace = _result["spec"]["jcloud"]["namespace"]

    stdout, stderr = cmd(f"kubectl get pods -n {namespace}")

    pods = []
    for i, line in enumerate(stdout.decode().split("\n")):
        if i == 0:
            continue
        cols = line.split()
        if len(cols) > 0:
            pod_name = cols[0]
            pods.append(pod_name)

    questions = [
        {
            'type': 'list',
            'name': 'pod',
            'message': 'Which pod do you want to check logs for?',
            'choices': pods,
        }
    ]
    pod = maybe_prompt_user(questions, 'pod', **kwargs)

    container = "gateway" if "gateway" in pod else "executor"
    cmd(f"kubectl logs {pod} -n {namespace} -c {container}", std_output=True)



from collections import defaultdict

from docarray import Document
from fastapi import APIRouter

from now.executor.gateway.bff.app.settings import user_input_in_bff
from now.executor.gateway.bff.app.v1.models.info import (
    CountRequestModel,
    CountResponseModel,
    EncoderToDataclassFieldsModsResponseModel,
    FieldNamesToDataclassFieldsResponseModel,
    TagsResponseModel,
)
from now.executor.gateway.bff.app.v1.models.shared import BaseRequestModel
from now.executor.gateway.bff.app.v1.routers.helper import jina_client_post

router = APIRouter()


@router.post('/tags')
async def get_tags(data: BaseRequestModel) -> TagsResponseModel:
    response = await jina_client_post(
        request_model=data,
        docs=Document(),
        endpoint='/tags',
        target_executor=r'\Aindexer\Z',
    )
    return TagsResponseModel(tags=response[0].tags['tags'])


@router.post('/count')
async def get_count(data: CountRequestModel) -> CountResponseModel:
    response = await jina_client_post(
        request_model=data,
        docs=Document(),
        endpoint='/count',
        target_executor=r'\Aindexer\Z',
        parameters={'limit': data.limit},
    )
    return CountResponseModel(number_of_docs=response[0].tags['count'])


@router.post('/field_names_to_dataclass_fields')
async def get_field_names_to_dataclass_fields() -> FieldNamesToDataclassFieldsResponseModel:
    return FieldNamesToDataclassFieldsResponseModel(
        field_names_to_dataclass_fields=user_input_in_bff.field_names_to_dataclass_fields
    )


@router.post('/encoder_to_dataclass_fields_mods')
async def get_index_fields_dict() -> EncoderToDataclassFieldsModsResponseModel:
    index_fields_dict = defaultdict(dict)
    for index_field_raw, encoders in user_input_in_bff.model_choices.items():
        index_field = index_field_raw.replace('_model', '')
        dataclass_field = user_input_in_bff.field_names_to_dataclass_fields[index_field]
        modality = user_input_in_bff.index_field_candidates_to_modalities[index_field]
        for encoder in encoders:
            index_fields_dict[encoder][dataclass_field] = modality
    return EncoderToDataclassFieldsModsResponseModel(
        encoder_to_dataclass_fields_mods=index_fields_dict
    )



from __future__ import annotations, print_function, unicode_literals

from docarray.typing import Image, Text, Video

from now.utils import BetterEnum

NOW_GATEWAY_VERSION = '0.0.4-fix-custom-gateway-remote-4'
NOW_PREPROCESSOR_VERSION = '0.0.124-refactor-no-batchwise-indexing-2'
NOW_ELASTIC_INDEXER_VERSION = '0.0.147-feat-integrate-elastic-manager-17'
NOW_AUTOCOMPLETE_VERSION = '0.0.11-refactor-custom-gateway-103'


class Apps(BetterEnum):
    SEARCH_APP = 'search_app'


class DialogStatus(BetterEnum):
    CONTINUE = 'continue'
    BREAK = 'break'
    SKIP = 'skip'


class DatasetTypes(BetterEnum):
    DEMO = 'demo'
    PATH = 'path'
    DOCARRAY = 'docarray'
    S3_BUCKET = 's3_bucket'
    ELASTICSEARCH = 'elasticsearch'


class Models(BetterEnum):
    CLIP_MODEL = 'encoderclip'
    SBERT_MODEL = 'encodersbert'


SUPPORTED_FILE_TYPES = {
    Text: ['txt', 'md'],
    Image: ['jpg', 'jpeg', 'png', 'bmp', 'tiff', 'tif'],
    Video: ['gif'],
}

FILETYPE_TO_MODALITY = {
    filetype: modality
    for modality, filetypes in SUPPORTED_FILE_TYPES.items()
    for filetype in filetypes
}
AVAILABLE_MODALITIES_FOR_SEARCH = [Text, Image, Video]
AVAILABLE_MODALITIES_FOR_FILTER = [Text]
NOT_AVAILABLE_MODALITIES_FOR_FILTER = [Image, Video]

BASE_STORAGE_URL = (
    'https://storage.googleapis.com/jina-fashion-data/data/one-line/datasets'
)

EXTERNAL_CLIP_HOST = 'encoderclip-pretty-javelin-3aceb7f2cd.wolf.jina.ai'

DEFAULT_FLOW_NAME = 'nowapi'
PREFETCH_NR = 10
NUM_FOLDERS_THRESHOLD = 100
MAX_DOCS_FOR_TESTING = 50
SURVEY_LINK = 'https://10sw1tcpld4.typeform.com/to/VTAyYRpR?utm_source=cli'

TAG_INDEXER_DOC_HAS_TEXT = '_indexer_doc_has_text'
ACCESS_PATHS = '@cc'
FLOW_STATUS = 'Serving'
DEMO_NS = 'now-example-{}'

MODALITY_TO_MODELS = {
    Text: [
        {'name': 'Clip', 'value': Models.CLIP_MODEL},
        {'name': 'Sbert', 'value': Models.SBERT_MODEL},
    ],
    Image: [{'name': 'Clip', 'value': Models.CLIP_MODEL}],
    Video: [{'name': 'Clip', 'value': Models.CLIP_MODEL}],
}

NOWGATEWAY_BFF_PORT = 8080



import json
import os
from collections import defaultdict
from typing import Dict, List, Type

from docarray import Document, DocumentArray
from docarray.dataclasses import is_multimodal

from now.common.detect_schema import (
    get_first_file_in_folder_structure_s3,
    get_s3_bucket_and_folder_prefix,
)
from now.constants import MAX_DOCS_FOR_TESTING, DatasetTypes
from now.data_loading.elasticsearch import ElasticsearchExtractor
from now.log import yaspin_extended
from now.now_dataclasses import UserInput
from now.utils import flatten_dict, get_chunk_by_field_name, sigmap


def load_data(
    user_input: UserInput, data_class=None, print_callback=print
) -> DocumentArray:
    """Based on the user input, this function will pull the configured DocumentArray dataset ready for the preprocessing
    executor.

    :param user_input: The configured user object. Result from the Jina Now cli dialog.
    :param data_class: The dataclass that should be used for the DocumentArray.
    :param print_callback: The callback function that should be used to print the status.
    :return: The loaded DocumentArray.
    """
    da = None
    if user_input.dataset_type in [DatasetTypes.DOCARRAY, DatasetTypes.DEMO]:
        print_callback('⬇  Pull DocumentArray dataset')
        da = _pull_docarray(user_input.dataset_name, user_input.admin_name)
        da = _update_fields_and_metadata(da, user_input)
    elif user_input.dataset_type == DatasetTypes.PATH:
        print_callback('💿  Loading files from disk')
        da = _load_from_disk(user_input=user_input, data_class=data_class)
    elif user_input.dataset_type == DatasetTypes.S3_BUCKET:
        print_callback('🗄  Loading files from S3')
        da = _list_files_from_s3_bucket(user_input=user_input, data_class=data_class)
    elif user_input.dataset_type == DatasetTypes.ELASTICSEARCH:
        print_callback('🔍  Loading data from Elasticsearch')
        da = _extract_es_data(user_input=user_input, data_class=data_class)
    da = set_modality_da(da)
    _add_metadata_to_chunks(da, user_input)
    if da is None:
        raise ValueError(
            f'Could not load DocumentArray dataset. Please check your configuration: {user_input}.'
        )
    if 'NOW_CI_RUN' in os.environ:
        da = da[:MAX_DOCS_FOR_TESTING]
    return da


def _add_metadata_to_chunks(da, user_input):
    dataclass_fields_to_field_names = {
        v: k for k, v in user_input.field_names_to_dataclass_fields.items()
    }
    for doc in da:
        for dataclass_field, meta_dict in doc._metadata['multi_modal_schema'].items():
            field_name = dataclass_fields_to_field_names.get(dataclass_field, None)
            if 'position' in meta_dict:
                get_chunk_by_field_name(doc, dataclass_field)._metadata[
                    'field_name'
                ] = field_name


def _update_fields_and_metadata(
    da: DocumentArray, user_input: UserInput
) -> DocumentArray:
    """Add selected index fields to da, add the tags, remove non-index chunks,
    and update multi modal schema."""
    if not da:
        return da
    all_fields = da[0]._metadata['multi_modal_schema'].keys()
    for doc in da:
        filtered_chunks = []
        filtered_chunk_names = []
        for field in all_fields:
            field_doc = get_chunk_by_field_name(doc, field)
            if field not in user_input.index_fields:
                if field_doc.blob or field_doc.tensor is not None:
                    continue
                doc.tags.update(
                    {
                        field: field_doc.content
                        if isinstance(field_doc.content, str)
                        else field_doc.uri
                    }
                )
            else:
                filtered_chunks.append(field_doc)
                filtered_chunk_names.append(field)
        doc.chunks = filtered_chunks
        # keep only the index fields in metadata
        doc._metadata['multi_modal_schema'] = {
            field: doc._metadata['multi_modal_schema'][field]
            for field in filtered_chunk_names
        }
        # Update the positions accordingly to access the chunks
        for position, field in enumerate(filtered_chunk_names):
            doc._metadata['multi_modal_schema'][field]['position'] = int(position)

    return da


def _pull_docarray(dataset_name: str, admin_name: str) -> DocumentArray:
    dataset_name = (
        admin_name + '/' + dataset_name if '/' not in dataset_name else dataset_name
    )
    try:
        docs = DocumentArray.pull(name=dataset_name, show_progress=True)
        if is_multimodal(docs[0]):
            return docs
        else:
            raise ValueError(
                f'The dataset {dataset_name} does not contain a multimodal DocumentArray. '
                f'Please check documentation https://docarray.jina.ai/fundamentals/dataclass/construct/'
            )
    except Exception:
        raise ValueError(
            'DocumentArray does not exist or you do not have access to it. '
            'Make sure to add user name as a prefix. Check documentation here. '
            'https://docarray.jina.ai/fundamentals/cloud-support/data-management/'
        )


def _extract_es_data(user_input: UserInput, data_class: Type) -> DocumentArray:
    query = {
        'query': {'match_all': {}},
        '_source': True,
    }
    es_extractor = ElasticsearchExtractor(
        query=query,
        index=user_input.es_index_name,
        user_input=user_input,
        data_class=data_class,
        connection_str=user_input.es_host_name,
    )
    extracted_docs = es_extractor.extract()
    return extracted_docs


def _load_from_disk(user_input: UserInput, data_class: Type) -> DocumentArray:
    """
    Loads the data from disk into multimodal documents.

    :param user_input: The user input object.
    :param data_class: The dataclass to use for the DocumentArray.
    """
    dataset_path = user_input.dataset_path.strip()
    dataset_path = os.path.expanduser(dataset_path)
    if os.path.isfile(dataset_path):
        try:
            da = DocumentArray.load_binary(dataset_path)
            if is_multimodal(da[0]):
                da = _update_fields_and_metadata(da, user_input)
                return da
            else:
                raise ValueError(
                    f'The file {dataset_path} does not contain a multimodal DocumentArray.'
                    f'Please check documentation https://docarray.jina.ai/fundamentals/dataclass/construct/'
                )
        except Exception:
            print(f'Failed to load the binary file provided under path {dataset_path}')
            exit(1)
    elif os.path.isdir(dataset_path):
        with yaspin_extended(
            sigmap=sigmap, text="Loading data from folder", color="green"
        ) as spinner:
            spinner.ok('🏭')
            docs = from_files_local(
                dataset_path,
                user_input.index_fields,
                user_input.field_names_to_dataclass_fields,
                data_class,
            )
            return docs
    else:
        raise ValueError(
            f'The provided dataset path {dataset_path} does not'
            f' appear to be a valid file or folder on your system.'
        )


def from_files_local(
    path: str,
    fields: List[str],
    field_names_to_dataclass_fields: Dict,
    data_class: Type,
) -> DocumentArray:
    """Creates a Multi Modal documentarray over a list of file path or the content of the files.

    :param path: The path to the directory
    :param fields: The fields to search for in the directory
    :param field_names_to_dataclass_fields: The mapping of the field names to the dataclass fields
    :param data_class: The dataclass to use for the document

    :return: A DocumentArray with the documents
    """

    file_paths = []
    for root, dirs, files in os.walk(path):
        file_paths.extend(
            [os.path.join(root, file) for file in files if not file.startswith('.')]
        )
    folder_generator = os.walk(path, topdown=True)
    current_level = folder_generator.__next__()
    folder_structure = 'sub_folders' if len(current_level[1]) > 0 else 'single_folder'
    if folder_structure == 'sub_folders':
        docs = create_docs_from_subdirectories(
            file_paths, fields, field_names_to_dataclass_fields, data_class
        )
    else:
        docs = create_docs_from_files(
            file_paths, fields, field_names_to_dataclass_fields, data_class
        )
    return DocumentArray(docs)


def create_docs_from_subdirectories(
    file_paths: List,
    fields: List[str],
    field_names_to_dataclass_fields: Dict,
    data_class: Type,
    path: str = None,
    is_s3_dataset: bool = False,
) -> List[Document]:
    """
    Creates a Multi Modal documentarray over a list of subdirectories.

    :param file_paths: The list of file paths
    :param fields: The fields to search for in the directory
    :param field_names_to_dataclass_fields: The mapping of the field names to the dataclass fields
    :param data_class: The dataclass to use for the document
    :param path: The path to the directory
    :param is_s3_dataset: Whether the dataset is stored on s3

    :return: The list of documents
    """

    docs = []
    folder_files = defaultdict(list)
    for file in file_paths:
        path_to_last_folder = (
            '/'.join(file.split('/')[:-1])
            if is_s3_dataset
            else os.sep.join(file.split(os.sep)[:-1])
        )
        folder_files[path_to_last_folder].append(file)
    for folder, files in folder_files.items():
        kwargs = {}
        tags_loaded_local = {}
        _s3_uri_for_tags = ''
        file_info = [
            _extract_file_and_full_file_path(file, path, is_s3_dataset)
            for file in files
        ]
        # first store index fields given as files
        for file, file_full_path in file_info:
            if file in fields:
                kwargs[field_names_to_dataclass_fields[file]] = file_full_path
        # next check json files that can also contain index fields, and carry on data
        for file, file_full_path in file_info:
            if file.endswith('.json'):
                if is_s3_dataset:
                    _s3_uri_for_tags = file_full_path
                    for field in data_class.__annotations__.keys():
                        if field not in kwargs.keys():
                            kwargs[field] = file_full_path
                else:
                    with open(file_full_path) as f:
                        json_data = flatten_dict(json.load(f))
                    for field, value in json_data.items():
                        if field in fields:
                            kwargs[field_names_to_dataclass_fields[field]] = value
                        else:
                            tags_loaded_local[field] = value
        doc = Document(data_class(**kwargs))
        if _s3_uri_for_tags:
            doc._metadata['_s3_uri_for_tags'] = _s3_uri_for_tags
        elif tags_loaded_local:
            doc.tags.update(tags_loaded_local)
        docs.append(doc)
    return docs


def create_docs_from_files(
    file_paths: List,
    fields: List[str],
    field_names_to_dataclass_fields: Dict,
    data_class: Type,
    path: str = None,
    is_s3_dataset: bool = False,
) -> List[Document]:
    """
    Creates a Multi Modal documentarray over a list of files.

    :param file_paths: List of file paths
    :param fields: The fields to search for in the directory
    :param field_names_to_dataclass_fields: The mapping of the files to the dataclass fields
    :param data_class: The dataclass to use for the document
    :param path: The path to the directory
    :param is_s3_dataset: Whether the dataset is stored on s3

    :return: A list of documents
    """
    docs = []
    for file in file_paths:
        kwargs = {}
        file, file_full_path = _extract_file_and_full_file_path(
            file, path, is_s3_dataset
        )
        file_extension = file.split('.')[-1]
        if (
            file_extension == fields[0].split('.')[-1]
        ):  # fields should have only one index field in case of files only
            kwargs[field_names_to_dataclass_fields[fields[0]]] = file_full_path
            docs.append(Document(data_class(**kwargs)))
    return docs


def _list_s3_file_paths(bucket, folder_prefix):
    """
    Lists the s3 file paths in an optimized way by finding the best level to use concurrent calls on
    in the file structure, using a threadpool.

    :param bucket: The s3 bucket used
    :param folder_prefix: The root folder prefix

    :return: A list of all s3 paths
    """
    # TODO bucket is not thread safe and outputs duplicate files for different prefixes
    # first_file = get_first_file_in_folder_structure_s3(bucket, folder_prefix)
    # structure_identifier = first_file[len(folder_prefix) :].split('/')
    # folder_structure = (
    #     'sub_folders' if len(structure_identifier) > 1 else 'single_folder'
    # )
    #
    # def get_level_order_prefixes(folder_prefix, level=1):
    #     """
    #     Gets the list of prefixes in a specific level. Levels are defined by the folder structure as follows:
    #     level_1/level_2/.../level_n/file.ext
    #
    #     :param folder_prefix: The current level prefix
    #     :param level: The desired level we want to get to
    #
    #     :return: A list of prefixes
    #     """
    #     level_prefixes = [
    #         obj['Prefix']
    #         for obj in bucket.meta.client.list_objects(
    #             Bucket=bucket.name, Prefix=folder_prefix, Delimiter='/'
    #         )['CommonPrefixes']
    #     ]
    #     if level == 1:
    #         return level_prefixes
    #     else:
    #         prefix_list = []
    #         for prefix in level_prefixes:
    #             prefix_list += get_level_order_prefixes(prefix, level - 1)
    #     return prefix_list
    #
    # def get_prefixes(max_levels=len(structure_identifier) - 2):
    #     """
    #     Finds the best level for the prefixes
    #
    #     :param max_levels: The maximum number of level we can get to, this defaults to len(structure_identifier) - 2
    #     because the latest level (len(structure_identifier) - 1) will only have files, so it won't have any common
    #     prefixes inside.
    #
    #     :return: A list of prefixes
    #     """
    #     level = 1
    #     list_prefixes = get_level_order_prefixes(folder_prefix, level)
    #     prefixes_states = [list_prefixes]
    #     while level < max_levels and len(list_prefixes) < NUM_FOLDERS_THRESHOLD:
    #         level += 1
    #         list_prefixes = get_level_order_prefixes(folder_prefix, level)
    #         prefixes_states.append(list_prefixes)
    #     if len(list_prefixes) > NUM_FOLDERS_THRESHOLD and len(prefixes_states) > 1:
    #         return prefixes_states[-2]
    #     return list_prefixes
    #
    # objects = []
    # if folder_structure == 'sub_folders':
    #     prefixes = get_prefixes()
    #     # TODO: change cpu count to a fixed number
    #     with ThreadPoolExecutor(max_workers=20) as executor:
    #         futures = []
    #         for prefix in prefixes:
    #             pref = ''.join(prefix)
    #             f = executor.submit(
    #                 lambda: list(bucket.objects.filter(Prefix=f'{pref}'))
    #             )
    #             futures.append(f)
    #         for f in futures:
    #             objects += f.result()
    # else:
    #     objects = list(bucket.objects.filter(Prefix=folder_prefix))
    objects = list(bucket.objects.filter(Prefix=folder_prefix))
    return [
        obj.key
        for obj in objects
        if not obj.key.endswith('/') and not obj.key.split('/')[-1].startswith('.')
    ]


def _list_files_from_s3_bucket(
    user_input: UserInput, data_class: Type
) -> DocumentArray:
    """
    Loads the data from s3 into multimodal documents.

    :param user_input: The user input object.
    :param data_class: The dataclass to use for the DocumentArray.

    :return: The DocumentArray with the documents.
    """
    bucket, folder_prefix = get_s3_bucket_and_folder_prefix(user_input)
    first_file = get_first_file_in_folder_structure_s3(bucket, folder_prefix)
    structure_identifier = first_file[len(folder_prefix) :].split('/')
    folder_structure = (
        'sub_folders' if len(structure_identifier) > 1 else 'single_folder'
    )
    with yaspin_extended(
        sigmap=sigmap, text="Listing files from S3 bucket ...", color="green"
    ) as spinner:
        file_paths = _list_s3_file_paths(bucket, folder_prefix)
        spinner.ok('🏭')

    with yaspin_extended(
        sigmap=sigmap, text="Creating docarray from S3 bucket files ...", color="green"
    ) as spinner:
        if folder_structure == 'sub_folders':
            docs = create_docs_from_subdirectories(
                file_paths,
                user_input.index_fields,
                user_input.field_names_to_dataclass_fields,
                data_class,
                user_input.dataset_path,
                is_s3_dataset=True,
            )
        else:
            docs = create_docs_from_files(
                file_paths,
                user_input.index_fields,
                user_input.field_names_to_dataclass_fields,
                data_class,
                user_input.dataset_path,
                is_s3_dataset=True,
            )
        spinner.ok('👝')
    return DocumentArray(docs)


def _extract_file_and_full_file_path(file_path, path=None, is_s3_dataset=False):
    """
    Extracts the file name and the full file path from s3 object.

    :param file_path: The file path
    :param path: The path to the directory
    :param is_s3_dataset: Whether the dataset is stored on s3

    :return: The file name and the full file path
    """
    if is_s3_dataset:
        file = file_path.split('/')[-1]
        file_full_path = '/'.join(path.split('/')[:3]) + '/' + file_path
    else:
        file_full_path = file_path
        file = file_path.split(os.sep)[-1]
    return file, file_full_path


def _get_modality(document: Document):
    """
    Detect document's modality based on its `modality` or `mime_type` attributes.

    :param document: The document to detect the modality for.
    """

    modalities = ['text', 'image', 'video']
    if document.modality:
        return document.modality
    mime_type_class = document.mime_type.split('/')[0]
    if document.mime_type == 'application/json':
        return 'text'
    if mime_type_class in modalities:
        return mime_type_class
    document.summary()
    raise ValueError(f'Unknown modality')


def set_modality_da(documents: DocumentArray):
    """
    Set document's modality based on its `modality` or `mime_type` attributes.

    :param documents: The DocumentArray to set the modality for.

    :return: The DocumentArray with the modality set.
    """
    for doc in documents:
        for chunk in doc.chunks:
            chunk.modality = chunk.modality or _get_modality(chunk)
    return documents



import pathlib
import time

from now.deployment.deployment import cmd


def setup_elastic_service(
    kubectl_path: str,
) -> str:
    """Setup ElasticSearch service and return a connection string to connect to the service with.

    :param kubectl_path: path to kubectl binary
    :return: connection string for connecting to the ElasticSearch service.
    """
    MAX_RETRIES = 20
    cur_dir = pathlib.Path(__file__).parent.resolve()
    cmd(
        f'{kubectl_path} create -f https://download.elastic.co/downloads/eck/2.4.0/crds.yaml'
    )
    cmd(
        f'{kubectl_path} apply -f https://download.elastic.co/downloads/eck/2.4.0/operator.yaml'
    )
    cmd(f'{kubectl_path} create ns nowapi')
    cmd(f'{kubectl_path} apply -f {cur_dir}/../deployment/elastic_kind.yml')
    num_retries = 0
    es_password, error_msg = '', b''
    while num_retries < MAX_RETRIES:
        es_password, error_msg = cmd(
            [
                kubectl_path,
                "get",
                "secret",
                "quickstart-es-elastic-user",
                "-o",
                "go-template='{{.data.elastic | base64decode}}'",
            ]
        )
        if es_password:
            es_password = es_password.decode("utf-8")[1:-1]
            break
        else:
            num_retries += 1
            time.sleep(2)
    if not es_password:
        raise Exception(error_msg.decode("utf-8"))
    host = f"https://elastic:{es_password}@quickstart-es-http.default:9200"
    return host



import os
from typing import Dict, List, Optional, Tuple, TypeVar

from docarray import DocumentArray
from jina import __version__ as jina_version

from now.app.base.create_jcloud_name import create_jcloud_name
from now.app.base.preprocess import preprocess_image, preprocess_text, preprocess_video
from now.constants import DEMO_NS, NOW_GATEWAY_VERSION
from now.demo_data import DemoDataset
from now.executor.name_to_id_map import name_to_id_map
from now.now_dataclasses import DialogOptions, UserInput


class JinaNOWApp:
    """
    Interface for Jina NOW applications.
    To create a new application, you need to inherit from this class and override the methods.
    Example:
        class TextToImageApp(JinaNowApp):
            pass
    """

    def __init__(self):
        self.flow_yaml = {}

    @property
    def app_name(self) -> str:
        """
        Name of the app. Should be an enum value set in now.constants.Apps
        """
        raise NotImplementedError()

    @property
    def is_enabled(self) -> bool:
        """
        Set to True if this app is enabled for the end user.
        """
        raise NotImplementedError()

    @property
    def description(self) -> str:
        """
        Short description of the app.
        """
        return 'Jina NOW app'

    @property
    def bff(self) -> Optional[str]:
        """
        TODO This function is currently not used but already introduces the concept of custom bff
        Used to configure the bff which is used to define input and output format.
        :return: the path to the file where the bff routs are configured
        """
        return None

    @property
    def playground(self) -> Optional[str]:
        """
        TODO This function is currently not used but already introduces the concept of custom playground
        Used to configure the playground(streamlit app) where the user can run example queries
        :return: the path to the streamlit file.
        """
        return None

    @property
    def options(self) -> List[DialogOptions]:
        """
        Get the options which are used to configure the app. Base class should override this function and
        return the list of their option. Check ``DialogOptions`` for the format of the options
        On CLI the user will get a prompt and at the storefront, a GUI will be generated accordingly.
        Example:
        ``return [options.your_custom_options`]``
        :return: List[DialogOptions]
        """
        return []

    @property
    def demo_datasets(self) -> Dict[TypeVar, List[DemoDataset]]:
        """Get a list of example datasets for the app."""
        raise NotImplementedError()

    @property
    def finetune_datasets(self) -> [Tuple]:
        """Defines the list of demo datasets which are fine-tunable."""
        return ()

    def set_app_parser(self, parser, formatter) -> None:
        """
        This parser reads from the `options` property and parses it
        to form the command line arguments for app
        """
        if self.is_enabled:
            parser = parser.add_parser(
                self.app_name,
                help=self.description,
                description=f'Create an {self.app_name} app.',
                formatter_class=formatter,
            )
            for option in self.options:
                if option.is_terminal_command:
                    parser.add_argument(
                        f'--{option.name}',
                        help=option.description,
                        type=str,
                    )

    def get_gateway_stub(self, user_input, testing=False) -> Dict:
        """Returns the stub for gateway in the flow."""
        gateway_stub = {
            'uses': f'jinahub+docker://{name_to_id_map.get("NOWGateway")}/{NOW_GATEWAY_VERSION}'
            if not testing
            else 'NOWGateway',
            'protocol': ['http', 'grpc'],
            'port': [8081, 8085],
            'monitoring': True,
            'cors': True,
            'uses_with': {'user_input_dict': user_input.to_safe_dict()},
            'env': {'JINA_LOG_LEVEL': 'DEBUG'},
            'jcloud': {
                'resources': {'instance': 'C5'},
                'capacity': 'spot',
            },
        }
        if 'NOW_EXAMPLES' in os.environ:
            gateway_stub['jcloud'] = {
                'custom_dns': [
                    f'{DEMO_NS.format(user_input.dataset_name.split("/")[-1])}.dev.jina.ai'
                ]
            }
        return gateway_stub

    def get_executor_stubs(self, user_input, testing=False, **kwargs) -> Dict:
        """
        Returns the stubs for the executors in the flow.
        """
        raise NotImplementedError()

    def setup(self, user_input: UserInput, testing=False, **kwargs) -> Dict:
        """Runs before the flow is deployed to setup the flow in self.flow_yaml.
        Common use cases:
            - create a database
            - finetune a model + push the artifact
            - notify other services
            - check if starting the app is currently possible
        :param user_input: user configuration based on the given options
        :param testing: use local executors if True
        """
        # Creates generic configuration such as labels in the flow
        # Keep this function as simple as possible. It should only be used to add generic configuration needed
        # for all apps. App specific configuration should be added in the app specific setup function.
        flow_yaml_content = {
            'jtype': 'Flow',
            'with': {
                'name': 'nowapi',
                'env': {'JINA_LOG_LEVEL': 'DEBUG'},
            },
            'jcloud': {
                'version': jina_version,
                'labels': {'team': 'now'},
                'name': create_jcloud_name(user_input.flow_name),
            },
            'gateway': self.get_gateway_stub(user_input, testing),
            'executors': self.get_executor_stubs(user_input, testing, **kwargs),
        }
        # Call the gateway stub function to get the gateway for the flow
        # Call the executor stubs function to get the executors for the flow
        # append user_input and api_keys to all executors except the remote executors
        user_input_dict = user_input.to_safe_dict()
        admin_emails = user_input.admin_emails or [] if user_input.secured else []
        user_emails = user_input.user_emails or [] if user_input.secured else []
        api_key = (
            [user_input.api_key] if user_input.secured and user_input.api_key else []
        )
        for executor in flow_yaml_content['executors']:
            if not executor.get('external', False):
                if not executor.get('uses_with', None):
                    executor['uses_with'] = {}
                executor['uses_with']['user_input_dict'] = user_input_dict
                executor['uses_with']['api_keys'] = api_key
                executor['uses_with']['user_emails'] = user_emails
                executor['uses_with']['admin_emails'] = admin_emails
        self.flow_yaml = self.add_telemetry_env(flow_yaml_content)

    def preprocess(
        self,
        docs: DocumentArray,
    ) -> DocumentArray:
        """Loads and preprocesses every document such that it is ready for indexing."""
        for doc in docs:
            for chunk in doc.chunks:
                try:
                    if chunk.modality == 'text':
                        preprocess_text(chunk)
                    elif chunk.modality == 'image':
                        preprocess_image(chunk)
                    elif chunk.modality == 'video':
                        preprocess_video(chunk)
                    else:
                        raise ValueError(f'Unsupported modality {chunk.modality}')
                except Exception as e:
                    chunk.summary()
                    print(e)
        return docs

    def is_demo_available(self, user_input) -> bool:
        raise NotImplementedError()

    @property
    def max_request_size(self) -> int:
        """Max number of documents in one request"""
        return 32

    @staticmethod
    def add_telemetry_env(flow_yaml_content):
        if 'JINA_OPTOUT_TELEMETRY' in os.environ:
            flow_yaml_content['gateway']['env']['JINA_OPTOUT_TELEMETRY'] = os.environ[
                'JINA_OPTOUT_TELEMETRY'
            ]
            for executor in flow_yaml_content['executors']:
                executor['env']['JINA_OPTOUT_TELEMETRY'] = os.environ[
                    'JINA_OPTOUT_TELEMETRY'
                ]
        return flow_yaml_content



from collections import defaultdict
from typing import Dict, Hashable, List, TypeVar

from docarray import Document, dataclass, field
from docarray.typing import Image, Video

from now.constants import AVAILABLE_MODALITIES_FOR_SEARCH, DatasetTypes
from now.now_dataclasses import UserInput
from now.utils import docarray_typing_to_modality_string


def update_dict_with_no_overwrite(dict1: Dict, dict2: Dict):
    """
    Update dict1 with dict2, but only if the key does not exist in dict1

    :param dict1: dict to be updated
    :param dict2: dict to be used for updating
    """
    for key, value in dict2.items():
        if key not in dict1:
            dict1[key] = value


def create_dataclass(
    fields: List = None,
    fields_modalities: Dict = None,
    dataset_type: DatasetTypes = None,
    user_input: UserInput = None,
):
    """
    Create a dataclass from the selected index fields
    and their corresponding modalities or directly from the user input which should
    contain that information. If both are provided, the user input will be used.

    for example:
    the index fields modalities can be:
    {'test.txt': Text , 'image.png': Image}

    the dataclass will be:

    @dataclass
    class DataClass:
        text_0: Text
        image_0: Image
        price: float
        description: str

    :param fields: list of fields
    :param fields_modalities: dict of fields and their modalities
    :param dataset_type: dataset type
    :param user_input: user inputs

    :return: dataclass object
    """

    if user_input:
        fields_modalities = user_input.index_field_candidates_to_modalities
        dataset_type = user_input.dataset_type
        fields = user_input.index_fields

    field_names_to_dataclass_fields = create_dataclass_fields_file_mappings(
        fields,
        fields_modalities,
    )
    (all_annotations, all_class_attributes,) = create_annotations_and_class_attributes(
        fields,
        fields_modalities,
        field_names_to_dataclass_fields,
        dataset_type,
    )
    mm_doc = type("MMDoc", (object,), all_class_attributes)
    setattr(mm_doc, '__annotations__', all_annotations)
    mm_doc = dataclass(mm_doc)

    return mm_doc, field_names_to_dataclass_fields


def create_annotations_and_class_attributes(
    fields: List,
    fields_modalities: Dict[str, TypeVar],
    field_names_to_dataclass_fields: Dict,
    dataset_type: DatasetTypes = None,
):
    """
    Create annotations and class attributes for the dataclass
    In case of S3 bucket, new field is created to prevent uri loading


    :param fields: list of fields
    :param fields_modalities: dict of fields and their modalities
    :param field_names_to_dataclass_fields: dict of selected field names and their corresponding fields in dataclass
    :param dataset_type: dataset type
    """
    annotations = {}
    class_attributes = {}
    ImageType, image_setter, image_getter = create_blob_type('Image')
    VideoType, video_setter, video_getter = create_blob_type('Video')

    for f in fields:
        if not isinstance(f, Hashable):
            continue
        if dataset_type == DatasetTypes.S3_BUCKET:
            S3Type, s3_setter, s3_getter = create_s3_type(str(fields_modalities[f])[1:])
            annotations[field_names_to_dataclass_fields[f]] = S3Type
            class_attributes[field_names_to_dataclass_fields[f]] = field(
                setter=s3_setter, getter=s3_getter, default=''
            )
        else:
            annotations[field_names_to_dataclass_fields[f]] = fields_modalities[f]
            if fields_modalities[f] == Image:
                class_attributes[field_names_to_dataclass_fields[f]] = field(
                    setter=image_setter, getter=image_getter, default=''
                )
            elif fields_modalities[f] == Video:
                class_attributes[field_names_to_dataclass_fields[f]] = field(
                    setter=video_setter, getter=video_getter, default=''
                )
            else:
                class_attributes[field_names_to_dataclass_fields[f]] = None
    return annotations, class_attributes


def create_s3_type(modality: str):
    """Create a new type for S3 bucket which sets the right modality"""
    S3Object = TypeVar('S3Object', bound=str)

    def my_setter(value) -> 'Document':
        """
        Custom setter for the S3Object type that doesn't load the content from the URI
        """
        doc = Document(uri=value)
        doc.modality = modality.lower()
        return doc

    def my_getter(doc: 'Document'):
        return doc.uri

    return S3Object, my_setter, my_getter


def create_blob_type(modality: str):
    """Creates a new type which loads into blob instead of tensor"""
    BlobObject = TypeVar(modality, bound=str)

    def my_setter(value) -> 'Document':
        """Custom setter for the BlobObject type that loads the content from the URI"""
        doc = Document(uri=value).load_uri_to_blob(timeout=10)
        doc.modality = modality.lower()
        return doc

    def my_getter(doc: 'Document'):
        return doc.uri

    return BlobObject, my_setter, my_getter


def create_dataclass_fields_file_mappings(fields: List, fields_modalities: Dict):
    """
    Create a mapping between the dataclass fields and the file fields

    :param fields: list of fields
    :param fields_modalities: dict of fields and their modalities
    """

    modalities_count = defaultdict(int)

    file_mapping_to_dataclass_fields = {}
    for f in fields:
        if not isinstance(f, Hashable):
            continue
        field_modality = fields_modalities[f]
        if field_modality in AVAILABLE_MODALITIES_FOR_SEARCH:
            file_mapping_to_dataclass_fields[
                f
            ] = f'{docarray_typing_to_modality_string(field_modality)}_{modalities_count[field_modality]}'
            modalities_count[fields_modalities[f]] += 1
    return file_mapping_to_dataclass_fields
